@@ -1,10 +1,11 @@
-#!/usr/bin/env python
+# coding: utf-8
+
+from __future__ import unicode_literals, division
 
 """
 This module implements basic kinds of jobs for QChem runs.
 """
 
-from __future__ import division
 import os
 from monty.io import zopen
 import shutil
